#include <iostream>
#include <vector>
#include <cstdio> 
#include <algorithm>
#include <set>
#include <string>
#include <map>
#include <fstream> 

using namespace std;

#define REP(i, to) for(int i=0; i<to; i++)
typedef pair<int, int> PII;

string concat(const vector<string>& s, string delim){
  string result;
  REP(i, s.size()){
    if(i != 0) result += delim; 
    result += s[i];
  }
  return result; 
}

int main()
{
  //pos-tag spans (i.e. topics) to id
  map<string, int> spans; 
  map<int, string> span_lookup; 
  map<int, int> span_counts; 
  
  //context (i.e. documents start word, end word) -> span_id (spans) -> count 
  map<string, map<int, int> > contexts; 
  map<string, int> context_counts; 
  map<string, int> context_id;
  
  int span_context_count = 0; 
  
  int nf; 
  while(cin >> nf){
    vector<string> tags(nf, ""); 
    vector<string> words(nf+2, ""); 
    
    REP(i, nf) {
      cin >> tags[i];
      cin >> words[i+1];
    }
    
    words[0]="SS";
    words[nf+1]="EE"; 
    
//    cout << "Words=" << concat(words, "|") << endl;
//    cout << "Tags=" << concat(tags, "|") << endl;
    
    REP(a, nf+2) for(int b=a+2; b<nf+2; b++){
      vector<string> sub(tags.begin()+a, tags.begin()+b-1);
      
//      cout << "Context " << words[a] << ":" << words[b] << "=" << concat(sub, " ") << endl;
  
      string context = words[a]+":"+words[b]; 
      if(context_id.count(context) == 0) context_id[context]=context_id.size()-1; 
  
      string span = concat(sub, ","); 
      if(spans.count(span) == 0) spans[span]=spans.size()-1; 
      
      int span_id = spans[span];
      span_lookup[span_id] = span; 
      span_counts[span_id]++; 
      contexts[context][span_id]++; 
      context_counts[context]++; 
      span_context_count++; 
    }
  }
 
  cout << "Probs" << endl;
  ofstream pout("probs.txt");
  vector<map<int, double> > dist(spans.size());
  int pi=0;   
  
  for(map<string, map<int, int> >::iterator ci = contexts.begin() ; ci != contexts.end() ; ci++){
    for(map<int,int>::iterator si = ci->second.begin() ; si != ci->second.end(); si++){
      double cc = context_counts[ci->first];
      double sc = span_lookup.size(); 
      pout << ci->first << "|" <<  span_lookup[si->first] << " " << ((double)si->second / cc) << " " << cc << endl;
      for(map<int,int>::iterator sj = ci->second.begin() ; sj != ci->second.end(); sj++){
        //dist[si->first][sj->first] += (double)(si->second * sj->second) / (cc * cc); //P(span1|context)P(span2|context)
        dist[si->first][sj->first] += ((double)(si->second * sj->second));
//        / ((double) (span_context_count * span_context_count)); //P(context|span1)P(context|span2)
      }
    }
    pi++;
    if(pi % 10000 == 0) cout << pi << endl;
  }
  pout.close(); 
  
  cout << "Spans" << endl;
  ofstream sout("spans.txt"); 
  REP(i, span_lookup.size()){
    sout << span_counts[i] << " " << span_lookup[i] << endl;
  }
  sout.close(); 
 
  cout << "Neighs" << endl;
  ofstream nout("neighs.txt");
  vector<pair<double, pair<int, int> > > N; 
   
  REP(i, dist.size()){
    if(i % 10000 == 0) cout << i << endl; 
    double scc = span_context_count; 
    for(map<int, double>::iterator iter = dist[i].begin(); iter != dist[i].end() ; iter++){
      if(iter->first > i){
        N.push_back(pair<double, pair<int, int> >((iter->second / scc) /scc, pair<int, int>(i, iter->first)));
      }
    }
  } 
  cout << "--Sort" << endl;
  sort(N.begin(), N.end(), greater<pair<double, pair<int, int> > >()); 
  
  cout << "--Save" << endl; 
  double sum=0.0; 
  REP(i, N.size()) {
    nout << N[i].first << " " << N[i].second.first << " " << N[i].second.second << endl;
    sum += N[i].first; 
    if(i%100000==0) cout << i << endl;
  }
  
  nout.close();
 
  cout << "Contexts: " << contexts.size() << endl;
  cout << "Spans: " << spans.size() << endl;
  cout << "Neighs: " << N.size() << " sum: " << sum << endl;
  
	return 0;
}
