package sned;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

//condition (isLeaf() XOR label != null)
//condition (isLeaf() || (lchild != null && rchild != null))

public class Constituent {
	private String label = null;

	private Constituent lchild = null;
	private Constituent rchild = null;
    
    private int length = -1;

	//Leaf
	public Constituent(String l){
		this.label = l;
		this.lchild = null;
		this.rchild = null;
        //this.length = 0;
	}

	//Node
	public Constituent(Constituent lch, Constituent rch){
		this.label = null;
		this.lchild = lch;
		this.rchild = rch;
        //this.length = lch.length() + rch.length();
	}
    
    //public int length() { return length; }
    public int length() {
        if (length < 0) {
            if (isLeaf()) length = 1;
            else length = lchild.length() + rchild.length();
        }
        return length;
    }

	public boolean isLeaf(){
		return (lchild == null && rchild == null);
	}

	//at least of length 2
	public List<String> getAllConstituentSpans(){
		List<String> result = new ArrayList<String>();
		if(!this.isLeaf()){
			result.addAll(lchild.getAllConstituentSpans());
			result.addAll(rchild.getAllConstituentSpans());
			result.add(this.toSpan());
		}
		return result;
	}
    
	public Set<String> getAllLabeledConstituentSpans(int pos){
		Set<String> result = new TreeSet<String>();
		if(!this.isLeaf()){
			result.addAll(lchild.getAllLabeledConstituentSpans(pos));
			result.addAll(rchild.getAllLabeledConstituentSpans(pos + lchild.length()));
			result.add(pos + ":" + this.toSpan());
		}
		return result;
	}

	public String toSpan(){
		if(this.isLeaf()) {
			return label;
		}
		return lchild.toSpan() + " " + rchild.toSpan();
	}

	static public List<Constituent> getAllParseTrees(SentenceSpan s){
		ArrayList<Constituent> result = new ArrayList<Constituent>();

		if(s.length()==0) {
			return null; //error
		}
		if(s.length()==1) {
			result.add(new Constituent(s.getLabel(0)));
		}
		//divide to <0,i) and <i, s.length())
		if(s.length()>=2){
			for(int i=1; i<s.length() ; i++){
				List<Constituent> lt = getAllParseTrees(s.getSpan(1, i));
				List<Constituent> rt = getAllParseTrees(s.getSpan(i+1, s.length()));

				for(Constituent lch : lt){
					for(Constituent rch : rt){
						result.add(new Constituent(lch, rch));
					}
				}
			}
		}

		return result;
	}


	public boolean equals(Object o) {
		if (o instanceof Constituent) {
			Constituent n = (Constituent)o;
			if(this.isLeaf()) {
				return n.isLeaf();
			}
			return lchild.equals(n.lchild) && rchild.equals(n.rchild);

			//NullPointerException if (lchild == null XOR rchild == null)
		}
		return false;
	}

	public String toString() {
		if(this.isLeaf()) {
			return label;
		}
		return "(" + lchild + " " + rchild + ")";
	}
}