package sned;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import topics.Topics;


public class CostFunction {

	String spansFileName = "data/spans.txt";
	String costsFileName = "data/neighs.txt";
    //String spanContextsFileName = "data/spans_with_contexts";
    
    Topics topics = null;//new Topics("data/wikidata_keys", "data/wikidata_weights", "data/wikidata_frequencies");

	private class Pair implements Comparable<Pair> {
		public int first;
		public int second;
		public Pair(int f, int s) { first = f; second = s; }
		public int hashCode() { return first << 16 + second; }
		public boolean equals(Object o) {
			if (o instanceof Pair) {
				Pair p = (Pair)o;
				return (first == p.first && second == p.second);
			}
			return false;
		}
		public int compareTo(Pair p) {
			if (this.first < p.first) return -1;
			if (this.first > p.first) return 1;

			if (this.second < p.second) return -1;
			if (this.second > p.second) return 1;

			return 0;
		}
	}

	Map<String, Integer> spanToIndex = new HashMap<String, Integer>();
	Map<Integer, String> indexToSpan = new TreeMap<Integer, String>();
	Map<Pair, Double> spanPairToCost = new TreeMap<Pair, Double>();

	public CostFunction() {
		loadSpans();
		loadCosts();
	}

	private void loadSpans() {
		System.out.println("Loading POS-tag spans...");
		try {
			BufferedReader in = new BufferedReader(new FileReader(spansFileName));
			String line;
			int index = 0;
			while ((line = in.readLine()) != null) {
			   String[] words = line.split(" ");
			   if (words.length >= 2) {
				   spanToIndex.put(words[1], index);
				   indexToSpan.put(index, words[1]);
				   ++index;
			   }
			}
			in.close();
		}
		catch (Exception e) {
			System.out.println("Failed to load POS-tag spans");
			e.printStackTrace();
		}
		System.out.println("Loaded " + spanToIndex.size() + " POS-tag spans");
	}

	private void loadCosts() {
		System.out.println("Loading POS-tag span substitution costs...");
		try {
			BufferedReader in = new BufferedReader(new FileReader(costsFileName));
			String line;
			int index = 0;
            double maxCost = Double.NEGATIVE_INFINITY;
            double minCost = Double.POSITIVE_INFINITY;
			while ((line = in.readLine()) != null) {
			   String[] words = line.split(" ");
			   if (words.length >= 3) {
				   int s1 = Integer.parseInt(words[1]);
				   int s2 = Integer.parseInt(words[2]);
				   double prob = Double.parseDouble(words[0]);
				   if(prob < 2.5e-14){ //memory limit in topic_neighs.txt
					   continue; 
				   }
				   double cost = 1/(0.01+ prob); //! minus (as cost file contains the probabilities)
                   //double cost = -Math.log(Double.parseDouble(words[0]));
				   spanPairToCost.put(new Pair(s1, s2), cost);
				   if (index % 100000 == 0) System.out.println(index);
				   ++index;
                   if (cost > maxCost) maxCost = cost;
                   if (cost < minCost) minCost = cost;
			   }
			   if(index > 8400000){ //memory limit in neighs.txt
				   break;
			   }
			}
			in.close();
            
            // rescalujeme costy do intervalu [0.5, 1.0]
            double desiredMinCost = 0.5;
            double desiredMaxCost = 1.0;
            double scale = (maxCost - minCost) / (desiredMaxCost - desiredMinCost);
            double offset = maxCost - desiredMaxCost * scale;
            
            /*double minCost2 = Double.POSITIVE_INFINITY;
            double maxCost2 = Double.NEGATIVE_INFINITY;
            Pair minPair2 = new Pair(0, 0);
            Pair maxPair2 = new Pair(0, 0);*/
            if (maxCost > Double.NEGATIVE_INFINITY) {
                for (Pair p : spanPairToCost.keySet()) {
                    spanPairToCost.put(p, (spanPairToCost.get(p) - offset) / scale);
                    /*if (minCost2 > spanPairToCost.get(p)) {
                        minCost2 = spanPairToCost.get(p);
                        minPair2 = p;
                    }
                    if (maxCost2 < spanPairToCost.get(p)) {
                        maxCost2 = spanPairToCost.get(p);
                        maxPair2 = p;
                    }*/
                }
            }
            /*System.out.println("minCost = " + minCost2);
            System.out.println(indexToSpan.get(minPair2.first));
            System.out.println(indexToSpan.get(minPair2.second));
            System.out.println("maxCost = " + maxCost2);
            System.out.println(indexToSpan.get(maxPair2.first));
            System.out.println(indexToSpan.get(maxPair2.second));*/
		}
		catch (Exception e) {
			System.out.println("Failed to load POS-tag span substitution costs");
			e.printStackTrace();
		}
		System.out.println("Loaded " + spanPairToCost.size() + " POS-tag spans");
	}

	public double getCost(SentenceSpan span1, SentenceSpan span2) {
		// rovnake spany maju cost 0.0
		if (span1.equals(span2)) return 0.0;

		Integer index1 = spanToIndex.get(span1.toString());
		Integer index2 = spanToIndex.get(span2.toString());

		//Integer index1 = span1.getIndex(spanToIndex);
		//Integer index2 = span2.getIndex(spanToIndex);

		// nezname spany maju cost... co ja viem...
		if (index1 == null || index2 == null) //return 1.0;
        return (span1.length() + span2.length()) / 2.0;
        //return span1.length();

		if (index1 > index2) {
			int tmp = index1;
			index1 = index2;
			index2 = tmp;
		}

		Double cost = spanPairToCost.get(new Pair(index1, index2));
        if (cost != null) return cost;
		//if (cost != null) return 0.5 * cost * (span1.length() + span2.length()) / 2.0;

		//System.err.println("CostFunction.getCost - tu sme nemali byt :P");
        //return 1.0;
        return (span1.length() + span2.length()) / 2.0;
        //return span1.length();
		//return 0.0000031415;
		//return Double.POSITIVE_INFINITY;
	}

	public double getTopicCost(Sentence span1, Sentence span2) {
		if (span1.equals(span2)) return 0.0;
        
        double dpLeft = 0.0;
        if (span1.leftContext() == null && span2.leftContext() == null) {
            dpLeft = 1.0;
        }
        if (span1.leftContext() != null && span2.leftContext() != null) {
            dpLeft = topics.getDotProduct(span1.leftContext(), span2.leftContext());
        }
        
        double dpRight = 0.0;
        if (span1.rightContext() == null && span2.rightContext() == null) {
            dpRight = 1.0;
        }
        if (span1.rightContext() != null && span2.rightContext() != null) {
            dpRight = topics.getDotProduct(span1.rightContext(), span2.rightContext());
        }
        
        return (1.0 - dpLeft) * (1.0 - dpRight) / 2.0;
    }
    
    

}
