package sned;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;


public class PCFGTree {
	public String label;
	public List<PCFGTree> childs;
    
    private int length = -1;

	public PCFGTree(){
		label = null;
		childs = new ArrayList<PCFGTree>();
	}
    
    public int length() {
        if (length < 0) {
            length = 0;
            if (childs.isEmpty()) {
                length = 1;
            }
            else for (PCFGTree child : childs) {
                length += child.length();
            }
        }
        return length;
    }

	public List<PCFGRule> getAllRules(){
		List<PCFGRule> result = new ArrayList<PCFGRule>();
		
		if(!childs.isEmpty()){
			PCFGRule rule = new PCFGRule(); 
			rule.left = label; 
			for(PCFGTree child : childs){
				rule.right.add(child.label); 
			}
			result.add(rule);
		}
		for(PCFGTree child : childs){
			result.addAll(child.getAllRules());
		}
		
		return result; 
	}
	
	//at least of length 2
	public List<String> getAllConstituentSpans(){
		List<String> result = new ArrayList<String>();
		for(PCFGTree tree : childs){
			result.addAll(tree.getAllConstituentSpans());
		}
		if(childs.size() > 1){
			result.add(this.toSpan());
		}
		return result;
	}
    
	public Set<String> getAllLabeledConstituentSpans(int pos){
		Set<String> result = new TreeSet<String>();
		if(childs.size() > 1){
			result.add(pos + ":" + this.toSpan());
		}
		for(PCFGTree tree : childs){
			result.addAll(tree.getAllLabeledConstituentSpans(pos));
            pos += tree.length();
		}
		return result;
	}

	public String toSpan(){
		if(this.childs.isEmpty()) {
			return label;
		}
		String result="";
		for(int i=0; i<childs.size() ; i++){
			result += ((i==0)?"":" ") + childs.get(i).toSpan(); //TODO n^2
		}
		return result;
	}

	public String toString() {
		if(this.childs.isEmpty()) {
			return label;
		}
		if(this.childs.size() == 1) {
			return this.childs.get(0).toString();
		}
		String result="(";
		for(int i=0; i<childs.size() ; i++){
			result += ((i==0)?"":" ") + childs.get(i); //TODO n^2
		}
		result += ")";
		return result;
	}

	static List<PCFGTree> parseCorpus(String filename){
		System.out.println("Loading corpus PCFG grammar...");
		List<PCFGTree> result = new ArrayList<PCFGTree>();

		try {
			BufferedReader in = new BufferedReader(new FileReader(filename));
			String line;
			while ((line = in.readLine()) != null) {
				PCFGTree tree = new PCFGTree();
				PCFGTree.parseLine(line, 0, tree);
				result.add(tree);
			}
			in.close();
		}
		catch (Exception e) {
			System.out.println("Failed to load PCFG grammar " + filename);
			e.printStackTrace();
		}
		System.out.println("Loaded " + result.size() + " exemplars");
		return result;
	}

	static int parseLine(String s, int pos, PCFGTree act){
		if(pos>=s.length() || pos<0) return -1;
		if(s.charAt(pos)==')') return pos+1;

		while(s.charAt(pos)=='('){
			pos++;
			int epos=s.indexOf(' ', pos);

			//cout << "name:'" << nr->name << "'" << endl;
			PCFGTree nr = new PCFGTree();
			nr.label = s.substring(pos, epos);
			act.childs.add(nr);

			pos = parseLine(s, epos+1, nr);
			if(pos < 0 || pos >= s.length()) return -1;
		}

		int epos=s.indexOf(')', pos);
		//act->word = s.substr(pos, epos-pos);
		//cout << "word:'" << act->word << "'" << endl;

		epos++;
		while(epos<s.length() && s.charAt(epos)==' ') epos++;

		return epos;
	}
}
