package sned;

import java.util.*;

//!TODO namapovat POStagy na cisla
public class SNED {

	private static int KNN = 20; //K nearest neighbours

	private List<PCFGTree> exemplars_pcfg = PCFGTree.parseCorpus("data/corpus_train_gold.txt");
	private LSH lsh = new LSH(exemplars_pcfg, 500, 10);

	private Corpus exemplar_corpus = new Corpus("data/corpus_train.txt");

	private CostFunction costFunc = new CostFunction();

	public int lsh_count = 0; 

	public SNED(){
		this.lsh.buildCache(this.exemplar_corpus.sentences);
	}

	public double getDistance(Sentence s1, Sentence s2) {
		if (s1.length() < s2.length()) return getDistance(s2, s1); // toto nam netreba asi tiez
		if (s1.length() == 0 && s2.length() == 0) return 0.0; // kvoli deleniu nulou pri normalizacii

		double[][][] table = new double[s1.length()+1][s2.length()+1][s1.length()+s2.length()+3];

		// pre istotu kym sa debugguje...
		for (int i = 0; i <= s1.length(); ++i) {
			for (int j = 0; j <= s2.length(); ++j) {
				for (int k = 0; k <= s1.length() + s2.length() + 1; ++k) {
					table[i][j][k] = Double.POSITIVE_INFINITY;
				}
			}
		}

		// toto uz neni az tak pre istotu
		for (int k = 1; k <= s1.length() + s2.length() + 1; ++k) {
			table[0][0][k] = Double.POSITIVE_INFINITY;
		}

		// prazdny string na prazdny string ma cenu 0
		table[0][0][0] = 0.0;
		//table[0][0][1] = Double.POSITIVE_INFINITY;

		// prvy riadok, empty na druhe slovo
		for (int j = 1; j <= s2.length(); ++j) {
			table[0][j][0] = Double.POSITIVE_INFINITY;
			for (int k = 1; k <= j; ++k) { // viac operacii ako dlzka slova nema zmysel
				table[0][j][k] = Double.POSITIVE_INFINITY;
				for (int j2 = 1; j2 <= j; ++j2) { // od j2 = k plusminus c, to sa doladi
					table[0][j][k] = Math.min(
							table[0][j][k],
							table[0][j2-1][k-1] + getCost(SentenceSpan.EMPTY, s2.getSpan(j2, j))
							);
				}
			}
			//table[0][j][j+1] = Double.POSITIVE_INFINITY; // treba, netreba?
		}

		// prvy stlpec, prve slovo na empty
		for (int i = 1; i <= s1.length(); ++i) {
			table[i][0][0] = Double.POSITIVE_INFINITY;
			for (int k = 1; k <= i; ++k) {
				table[i][0][k] = Double.POSITIVE_INFINITY;
				for (int i2 = 1; i2 <= i; ++i2) {
					table[i][0][k] = Math.min(
							table[i][0][k],
							table[i2-1][0][k-1] + getCost(s1.getSpan(i2, i), SentenceSpan.EMPTY)
							);
				}
			}
			//table[i][0][i+1] = Double.POSITIVE_INFINITY;
		}

		// zvysok tabulky
		for (int i = 1; i <= s1.length(); ++i) {
			for (int j = 1; j <= s2.length(); ++j) {
				table[i][j][0] = Double.POSITIVE_INFINITY;
				for (int k = 1; k <= i+j; ++k) { // viac operacii ako sucet dlzok slov zase nema zmysel
					table[i][j][k] = Double.POSITIVE_INFINITY;
					for (int i2 = 1; i2 <= i; ++i2) {
						for (int j2 = 1; j2 <= j; ++j2) {
							//System.out.println(i + " " + j + " " + k);
							table[i][j][k] = min(
									table[i][j][k],
									table[i2-1][j][k-1] + getCost(s1.getSpan(i2, i), SentenceSpan.EMPTY),
									table[i][j2-1][k-1] + getCost(SentenceSpan.EMPTY, s2.getSpan(j2, j)),
									table[i2-1][j2-1][k-1] + getCost(s1.getSpan(i2, i), s2.getSpan(j2, j))
									);
						}
					}
				}
				//table[i][j][i+j+1] = Double.POSITIVE_INFINITY;
			}
		}

		double result = Double.POSITIVE_INFINITY;
		for (int k = 1; k <= s1.length() + s2.length(); ++k) {
			//System.out.println(table[s1.length()][s2.length()][k] + " in " + k + " steps");
			result = Math.min(result, table[s1.length()][s2.length()][k] / (double)k);
		}

		return result;
	}

	private double min(double d1, double d2, double d3, double d4) {
		return Math.min(Math.min(d1, d2), Math.min(d3, d4));
	}

	public double getCost(SentenceSpan s1, SentenceSpan s2) {
		//System.out.println(s1 + " | " + s2 + " | " + s1.equals(s2));
		// Rovnake spany maju cenu 0, toto je po novom CostFunction
		//if (s1.equals(s2)) return 0.0;

		// Mazanie a vkladanie zatial nepodporujeme
		if (s1.length() == 0 || s2.length() == 0) return Double.POSITIVE_INFINITY;

		return costFunc.getCost(s1, s2);
		//return costFunc.getTopicCost(s1, s2);
		//return costFunc.getCost(s1, s2) + 0.1 * costFunc.getTopicCost(s1, s2);
	}

	public Constituent searchNeighbors(Sentence s1) {
		System.out.println("searchNeighbors(" + s1 + ")");

		ArrayList<SentenceCost> neighs = new ArrayList<SentenceCost>();

		//WITHOUT LSH
		/*
		for (int i=0; i<exemplar_corpus.sentences.size() ; i++) {
			Sentence s2 = exemplar_corpus.sentences.get(i);
			Double dist = getDistance(s1, s2);
			neighs.add(new SentenceCost(dist, s2, exemplars_pcfg.get(i)));
		}
		/**/

		//PURE LSH (15,3,7,48,60 - collisions of first sentences)
		/**/
		Set<Integer> sentence_ids = new HashSet<Integer>(this.lsh.getCacheCollisions(s1));
		for (Integer sentence_id : sentence_ids) {
			Sentence s2 = exemplar_corpus.sentences.get(sentence_id);
			Double dist = getDistance(s1, s2);
			neighs.add(new SentenceCost(dist, s2, exemplars_pcfg.get(sentence_id)));
		}
		System.out.println("neighs.size()=" + neighs.size());
		if(neighs.size() < 30){
			return null; 
		}
		/**/
		
		//COMBINE LSH and WIHTOUT LSH if not enough exemplars found, then consider all
		/*
		if(neighs.size()<1){
			neighs.clear();
			for (int i=0; i<exemplar_corpus.sentences.size() ; i++) {
				Sentence s2 = exemplar_corpus.sentences.get(i);
				Double dist = getDistance(s1, s2);
				neighs.add(new SentenceCost(dist, s2, exemplars_pcfg.get(i)));
			}
		}
		/**/

		Collections.sort(neighs);
		int knn = Math.min(KNN, neighs.size());
		this.lsh_count = neighs.size();

		for(int i=0; i<10; i++){
			//System.out.println(neighs.get(i).cost + " <- " + s1 + " vs " + neighs.get(i).sentence + " ("+i+")" + neighs.get(i).gold);
		}
		for(int i=10; i<=100; i+=10){
			//System.out.println(neighs.get(i).cost + " <- " + s1 + " vs " + neighs.get(i).sentence + " ("+i+")" + neighs.get(i).gold);
		}

		//k nearest neigbours
		List<String> neigh_con = new ArrayList<String>();
		for(int i=0; i<knn; i++){
			neigh_con.addAll(neighs.get(i).gold.getAllConstituentSpans());
		}

		//System.out.println("neigh_con:");
		for(String neigh_span : neigh_con){
			//System.out.println(neigh_span);
		}

		List<Constituent> parse_trees = Constituent.getAllParseTrees(s1.getSpan());
		Constituent best_t=null;
		int best_v=-1;
		for(Constituent tree : parse_trees){
			List<String> tree_con = tree.getAllConstituentSpans();
			//System.out.println("tree_con " + tree);

			int v=0;
			for(String tree_span : tree_con){
				for(String neigh_span : neigh_con){
					if(neigh_span.equals(tree_span)){
						v++;
					}
				}
				//System.out.println(tree_span + " count=" + tc);
			}
			if(v>best_v){
				best_t = tree;
				best_v = v;
			}

			//System.out.println("value="+v);
		}

		System.out.println("Sentence:  " + s1.toWordsString());
		System.out.println("POStags:   " + s1);
		System.out.println("Best tree: " + best_t + " [" + best_v + "]");
		return best_t;
	}

	public String evaluate(Constituent treeTest, PCFGTree treeGold, Evaluator eval) {
		Set<String> tree_con1 = treeTest.getAllLabeledConstituentSpans(0);
		//for (String s : tree_con1) {
		//    System.out.println(s);
		//}
		Set<String> tree_con2 = treeGold.getAllLabeledConstituentSpans(0);
		//for (String s : tree_con2) {
		//    System.out.println(s);
		//}

		int sizeTest = tree_con1.size();
		int sizeGold = tree_con2.size();

		tree_con1.retainAll(tree_con2);
		int truePositives = tree_con1.size();

		// nezaratame matchnutu celu vetu
		if (sizeGold > 1 && sizeTest > 1) {
			truePositives--;
			sizeTest--;
			sizeGold--;
		}

		double precision = (double)truePositives / (double)sizeTest;
		double recall = (double)truePositives / (double)sizeGold; //!TODO can it be bigger than one ? 

		double f1 = 2 * precision * recall / (precision + recall);

		eval.addTestResults(truePositives, sizeGold, sizeTest);

		System.out.println("Precision: " + precision);
		System.out.println("Recall:    " + recall);
		System.out.println("F1 score:  " + f1);

		return precision + "\t" + recall + "\t" + f1; 
	}
}
