package sned;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;


public class Sentence {

	protected ArrayList<String> words = new ArrayList<String>();
	protected ArrayList<String> labels;// = new ArrayList<String>();
    protected String lContext = null, rContext = null;
    
    protected SentenceSpan[][] spans = null;

	public final static Sentence EMPTY = new Sentence();

	public Sentence() {
		labels = new ArrayList<String>();
		words = new ArrayList<String>();
        lContext = null;
        rContext = null;
	}

	public Sentence(String labels) {
		this("", labels);
	}

	public Sentence(String words, String labels) {
		this();
		if (labels.equals("")) this.labels = new ArrayList<String>();
		else this.labels = new ArrayList(Arrays.asList(labels.split(" ")));

		if (words.equals("")) this.words = new ArrayList<String>();
		else this.words = new ArrayList(Arrays.asList(words.split(" ")));
	}

	/*public Sentence(List<String> labels) {
		this.labels = new ArrayList<String>();
		this.labels.addAll(labels);
	}*/

	public Sentence(List<String> words, List<String> labels) {
		this(words, labels, null, null);
	}

	public Sentence(List<String> words, List<String> labels, String lContext, String rContext) {
		this.labels = new ArrayList<String>();
		this.labels.addAll(labels);

		this.words = new ArrayList<String>();
		this.words.addAll(words);
        
        this.lContext = lContext;
        this.rContext = rContext;
	}

	public static Sentence fromCorpusLine(String corpusLine) {
		Sentence ret = new Sentence();
		String[] words = corpusLine.split(" ");
		int num = Integer.parseInt(words[0]);
		for (int i = 0; i < num; ++i) {
			ret.labels.add(words[2*i+1]);
			ret.words.add(words[2*i+2]);
		}
		return ret;
	}

	//public Sentence getSpan(int i, int j) {
	//	//System.out.println(this + " | sublist(" + i + "," + j + ") | " + new Sentence(tags.subList(i, j)));
    //    String lContext2 = (i > 1) ? words.get(i-2) : null;
    //    String rContext2 = (j < words.size()) ? words.get(j) : null;
	//	return new Sentence(words.subList(i-1, j), labels.subList(i-1, j), lContext2, rContext2);
	//}
    
    public SentenceSpan getSpan() {
        return new SentenceSpan(this, 1, length());
    }
    
    public SentenceSpan getSpan(int i, int j) {
        if (spans == null) {
            spans = new SentenceSpan[length()][length()];
        }
        if (spans[i-1][j-1] == null) {
            spans[i-1][j-1] = new SentenceSpan(this, i, j);
        }
        return spans[i-1][j-1];
    }

	public String getLabel(int i){
		return labels.get(i);
	}

	public int length() {
		return labels.size();
	}
    
    public ArrayList<String> labels() {
        return labels;
    }
    
    public ArrayList<String> words() {
        return words;
    }
    
    public String leftContext() {
        return lContext;
    }
    
    public String rightContext() {
        return rContext;
    }

	public boolean equals(Object o) {
		if (o instanceof Sentence) {
			Sentence s = (Sentence)o;
			if (s.length() == this.length()) {
				for (int i = 0; i < this.length(); ++i) {
					if (!this.labels.get(i).equals(s.labels.get(i))) {
						return false;
					}
				}
				return true;
			}
		}
		return false;
	}

    public String toWordsString() {
		// vyrobi nieco taketo "PP,NN,VBS,NNS"
		if (words == null) return "null";
		if (words.size() == 0) return "";
		String result = words.get(0);
		for (int i = 1; i < words.size(); ++i) {
			result += " " + words.get(i);
		}
		return result;
    }

	public String toString() {
		// vyrobi nieco taketo "PP,NN,VBS,NNS"
		if (labels == null) return "null";
		if (labels.size() == 0) return "";
		String result = labels.get(0);
		for (int i = 1; i < labels.size(); ++i) {
			result += "," + labels.get(i);
		}
		return result;
	}

	public List<Sentence> alignDiff(Sentence other) {
		// TODO Auto-generated method stub
		return null;
	}
}
