
package utils;

import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

public class Counter<T> {

    protected double defaultValue = 0.0;
    protected double defaultStep = 1.0;
    protected Map<T, Double> map = new TreeMap<T, Double>();

    public Counter() {}

    public void setDefaultValue(double value) {
        defaultValue = value;
    }

    public void setDefaultStep(double value) {
        defaultStep = value;
    }

    public void add(T key) {
        add(key, defaultStep);
    }

    public void add(T key, double value) {
        if (!map.containsKey(key)) {
            map.put(key, defaultValue + value);
        }
        map.put(key, map.get(key) + value);
    }

    public void set(T key, double value) {
        map.put(key, value);
    }

    public double get(T key) {
        if (map.containsKey(key)) {
            return map.get(key);
        }
        return defaultValue;
    }

    public Set<T> keySet() {
        return map.keySet();
    }
}
