
package utils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.Scanner;

public class FileCleaner {
    private static String rawFolder = "../wikidata/raw/";
    private static String tmpFolder = "../wikidata/tmp/";
    private static String cleanFolder = "../wikidata/clean/";
    private static String countFile = "../wikidata/wikidata_frequencies/";
    private static int minFrequency = 10;

    private static Counter<String> counts = new Counter<String>();

    public static void cleanFiles() {
        try {
            System.out.println("Cleaning stage 1...");
            File dir1 = new File(rawFolder);
            for (File entry : dir1.listFiles()) {
                if (entry.isFile()) {
                    //System.out.println(entry.getName());
                    cleanFile1(entry.getName());
                }
            }
            System.out.println("Cleaning stage 2...");
            File dir2 = new File(rawFolder);
            for (File entry : dir2.listFiles()) {
                if (entry.isFile()) {
                    //System.out.println(entry.getName());
                    cleanFile2(entry.getName());
                }
            }
            saveFrequencies();
            System.out.println("Cleaning done");
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void cleanFile1(String fileName) throws Exception {
        Scanner in = new Scanner(new File(rawFolder + "/" + fileName));
        BufferedWriter out = new BufferedWriter(new FileWriter(tmpFolder + "/" + fileName));
        while (in.hasNext()) {
            String word = in.next();
            word = word.toLowerCase();
            if (word.matches("[&+/\\\\]")) continue;
            String words[] = word.split("[.,-_]");
            for (String word2 : words) {
                if (word2.matches("[a-zA-Z]+")) {
                    //System.out.println(word2);
                    counts.add(word2);
                    out.write(word2 + " ");
                }
            }
        }
        out.close();
        in.close();
    }

    public static void cleanFile2(String fileName) throws Exception {
        Scanner in = new Scanner(new File(tmpFolder + "/" + fileName));
        BufferedWriter out = new BufferedWriter(new FileWriter(cleanFolder + "/" + fileName));
        while (in.hasNext()) {
            String word = in.next();
            if (counts.get(word) >= minFrequency) {
                out.write(word + " ");
            }
        }
        out.close();
        in.close();
    }

    public static void saveFrequencies() throws Exception {
        BufferedWriter out = new BufferedWriter(new FileWriter(countFile));
        for (String word : counts.keySet()) {
            if (counts.get(word) > minFrequency) {
                out.write(word + "\t" + counts.get(word) + "\n");
            }
        }
        out.close();
    }

}
