#include <iostream>
#include <vector>
#include <cstdio> 
#include <algorithm>
#include <set>
#include <string>
#include <map>
//#include <hash_map>
#include <fstream> 

using namespace std;
using namespace __gnu_cxx;

#define REP(i, to) for(int i=0; i<to; i++)
typedef pair<int, int> PII;
typedef unsigned short ushort; 
typedef unsigned int uint; 

string concat(const vector<string>& s, string delim){
  string result;
  REP(i, s.size()){
    if(i != 0) result += delim; 
    result += s[i];
  }
  return result; 
}

int main()
{
  cout << "Loading topic_context_prob" << endl;

  map<string, int> dict;
  map<int, map<int, float> > dict_neighs; 
  map<int, string> dict_lookup; 

  ifstream topin("topic_context_prob"); 
  float tprob;
  string tw1, tw2;
  while(topin >> tprob >> tw1 >> tw2){
    if(dict.count(tw1)==0) dict[tw1]=dict.size()-1; 
    if(dict.count(tw2)==0) dict[tw2]=dict.size()-1;
    dict_neighs[dict[tw1]][dict[tw2]]=tprob;
    dict_lookup[dict[tw1]]=tw1; 
  }
  dict["SS"]=dict.size()-1; 
  dict["EE"]=dict.size()-1; 
  dict_neighs[dict["SS"]][dict["SS"]]=1.0;
  dict_neighs[dict["EE"]][dict["EE"]]=1.0;
  dict_lookup[dict["SS"]] = "SS";
  dict_lookup[dict["EE"]] = "EE";
  topin.close(); 

  cout << "Loaded " << dict_neighs.size() << " dict_neighs instances and " << dict.size() << " dict instances" << endl;

  //pos-tag spans (i.e. topics) to id
  map<string, ushort> spans; 
  map<int, string> span_lookup; 
  map<int, int> span_counts; 
  
  //context (i.e. documents start word, end word) -> span_id (spans) -> count 
  //word count ~43K => int + corpus words! 
  //span count ~33K => int
  map<uint, map<ushort, int> > contexts; 
  map<int, int> context_counts; 
  
  long long int span_context_count = 0; 
  
  cout << "Loading corpus" << endl;
  int corp_line_i=0; 
  
  int nf;  
  while(cin >> nf){
    vector<string> tags(nf, ""); 
    vector<string> words(nf+2, ""); 
    
    REP(i, nf) {
      cin >> tags[i];
      cin >> words[i+1];
    }
    
    words[0]="SS";
    words[nf+1]="EE"; 
    
    //each span
    REP(a, nf+2) for(int b=a+2; b<nf+2; b++){
      vector<string> sub(tags.begin()+a, tags.begin()+b-1);
  
      string span = concat(sub, ","); 
      if(spans.count(span) == 0) spans[span]=spans.size()-1; 
          
      ushort span_id = spans[span];
      span_lookup[span_id] = span; 
          
      if(dict.count(words[a]) == 0) {
        dict[words[a]]=dict.size()-1; 
        dict_neighs[dict[words[a]]][dict[words[a]]]=1.0; 
      }
        dict_lookup[dict[words[a]]] = words[a];
      if(dict.count(words[b]) == 0) {
        dict[words[b]]=dict.size()-1; 
        dict_neighs[dict[words[b]]][dict[words[b]]]=1.0; 
        dict_lookup[dict[words[b]]] = words[b];
      }
  
      map<int, float>::iterator iterend=dict_neighs[dict[words[a]]].end(); 
      map<int, float>::iterator jterend=dict_neighs[dict[words[b]]].end(); 
      
      for(map<int, float>::iterator iter=dict_neighs[dict[words[a]]].begin() ; iter != iterend ; iter++){
        for(map<int, float>::iterator jter=dict_neighs[dict[words[b]]].begin() ; jter != jterend ; jter++){
          //TODO probabilities of dict
          //string context = dict_lookup[iter->first]+":"+dict_lookup[jter->first]; 
          //pair<int,int> context = PII(iter->first, jter->first);  
          uint context = iter->first * 65536 + jter->first; 
      
          //cout << iter->first << ":" << jter->first << " " << context << " size=" << dict.size() << endl;
      
          contexts[context][span_id]++; 
          span_counts[span_id]++; 
          context_counts[context]++; 
          span_context_count++; 
        }
      }
    }
    if(++corp_line_i % 50 == 0) cout << "Loaded " << corp_line_i << " lines" << endl;  
  }
 
  cout << "Probs contexts("<<contexts.size()<<")" << endl;
  //ofstream pout("topic_probs.txt");
  vector<map<int, float> > dist(spans.size());
  int pi=0;   
  
  for(map<uint, map<ushort, int> >::iterator ci = contexts.begin() ; ci != contexts.end() ; ci++){
    vector<pair<int, int> > con_span_count; 
    for(map<ushort,int>::iterator si = ci->second.begin() ; si != ci->second.end(); si++){
      con_span_count.push_back(PII(si->first, si->second));
    }
    
    REP(si, con_span_count.size()){
      //float cc = context_counts[ci->first];
      //float sc = span_lookup.size(); 
      //pout << ci->first << "|" <<  span_lookup[si->first] << " " << ((float)si->second / cc) << " " << cc << endl;
      REP(sj, con_span_count.size()){
        dist[con_span_count[si].first][con_span_count[sj].first] += ((float)con_span_count[si].second * (float)con_span_count[sj].second);
        //cout << span_lookup[si->first] << " vs " << span_lookup[sj->first] << " + " << ((float)(si->second * sj->second)) << endl;
      }
    }
    pi++;
    if(pi % 50000 == 0) cout << pi << endl; //100000 -> 2min on whole corpus => ETA 208min ; true story is < 15min
  }
  //pout.close(); 
  cout << "Prob contexts done dist.size() = " << dist.size() << endl;
  
  cout << "Spans("<<span_lookup.size()<<")" << endl;
  ofstream sout("topic_spans.txt"); 
  REP(i, span_lookup.size()){
    sout << span_counts[i] << " " << span_lookup[i] << endl;
  }
  sout.close(); 
 
 /*
  cout << "Neighs dist("<<dist.size()<<")" << endl;
  ofstream nout("topic_neighs.txt");
  vector<pair<float, pair<int, int> > > N; 
   
  REP(i, dist.size()){
    if(i % 10000 == 0) cout << i << endl; 
    float scc = span_context_count; 
    for(map<int, float>::iterator iter = dist[i].begin(); iter != dist[i].end() ; iter++){
      if(iter->first > i){
        N.push_back(pair<float, pair<int, int> >((iter->second / scc) /scc, pair<int, int>(i, iter->first)));
      }
    }
  } 
  cout << "--Sort(" << N.size() << ")" << endl;
  sort(N.begin(), N.end(), greater<pair<float, pair<int, int> > >()); 
  
  cout << "--Save" << endl; 
  float sum=0.0; 
  REP(i, N.size()) {
    nout << N[i].first << " " << N[i].second.first << " " << N[i].second.second << endl;
    sum += N[i].first; 
    if(i%100000==0) cout << i << endl;
  }*/
  
  cout << "Saveing dist("<<dist.size()<<")" << endl;
  ofstream nout("topic_neighs.txt");
  int ncount=0; 
  double sum=0.0;
  REP(i, dist.size()){
    if(i % 25000 == 0) cout << i << endl; 
    double scc = span_context_count; 
    for(map<int, float>::iterator iter = dist[i].begin(); iter != dist[i].end() ; iter++){
      if(iter->first > i){
        //N.push_back(pair<float, pair<int, int> >((iter->second / scc) /scc, pair<int, int>(i, iter->first)));
        nout << (((double)iter->second / scc) /scc) << " " << i << " " << iter->first << endl;
        ncount++; 
        sum+=(((double)iter->second / scc) /scc); 
      }
    }
  } 
  
  nout.close();
 
  cout << "Words: " << dict.size() << endl;
  cout << "Contexts: " << contexts.size() << endl;
  cout << "Spans: " << spans.size() << endl;
  cout << "Neighs: " << ncount << " sum: " << sum << endl;
  
	return 0;
}
